// ArmDisassembler.cpp: implementation of the CArmDisassembler class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "ArmDisassembler.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CArmDisassembler::CArmDisassembler()
{
	TRACE("construct CArmDisassembler \n");
	
	// ??? all debugging stuff - nothing needs constructing

	/*	
	// open report file
	CStdioFile dissReportFile;
	BOOL successfulReportOpen = dissReportFile.Open(
		_T("C:\\Work\\Project\\Diss.txt"),
		CFile::modeCreate		|
		CFile::modeWrite		|
		CFile::shareDenyWrite	|
		CFile::typeText
		);

	uint8 testMemory[0x8000 + 200000];

	
	// load file into test memory
	CFile dissTestFile;
	CString fileName = "C:\\Work\\Project\\SourcePrograms\\dissTests\\StrongEdRI,ff8";
	uint bytesRead;
	if( dissTestFile.Open(fileName, CFile::modeRead) )
	{
		bytesRead = dissTestFile.Read(testMemory + 0x8000, 200000);
		TRACE("BytesRead from test file = %d\n", bytesRead);
		dissTestFile.Close();
	}
	else
	{
		TRACE(">>>>>>>>> DISS test file %s not found!\n", fileName);
	}
	
	for(uint32 address = 0x8000; address < 0x8000 + bytesRead; address += 4)
	{
		uint32 dissInstr = testMemory[address] | (testMemory[address+1]<<8);
		dissInstr |= (testMemory[address+2]<<16) | (testMemory[address+3]<<24);
		CString disassembly = disassemble(address, dissInstr) + "\n";
		CString addressS;
		addressS.Format("0x%x", address );
		disassembly = addressS + " : " + disassembly;
		dissReportFile.WriteString(disassembly);
	}
	*/
	
}

CArmDisassembler::~CArmDisassembler()
{

}

CString CArmDisassembler::disassemble(uint32 address, uint32 instruction)
{

	// decode based on bits 24 - 27 of instruction
	switch( getField(instruction, 24, 27) )
	{
		case 0	:	return decodeMultiplyOrDataProcessing(address, instruction);
		case 1	:	return decodeSingleDataSwapOrDataProcessing(address, instruction);
		case 2	:	return decodeDataProcessing(address, instruction);
		case 3	:	return decodeDataProcessing(address, instruction);
		case 4	:	return decodeSingleDTImmOffsetPostIndex(address, instruction);
		case 5	:	return decodeSingleDTImmOffsetPreIndex(address, instruction);
		case 6	:	return decodeSingleDTRegOffsetPostIndex(address, instruction);
		case 7	:	return decodeSingleDTRegOffsetPreIndex(address, instruction);
		case 8	:	return decodeBlockDTPostIndex(address, instruction);
		case 9	:	return decodeBlockDTPreIndex(address, instruction);
		case 10	:	return decodeBranch(address, instruction);
		case 11	:	return decodeBranchWithLink(address, instruction);
		case 12 :	return decodeCoProDTPreIndex(address, instruction);
		case 13 :	return decodeCoProDTPostIndex(address, instruction);
		case 14	:	return decodeCoProRegTransferOrDataOperation(address, instruction);
		case 15	:	return decodeSoftwareInterrupt(address, instruction);
	}

	return "ERROR DISASSEMBLING IN Disassemble()";
}

CString CArmDisassembler::decodeMultiplyOrDataProcessing(uint32 address, uint32 instruction)
{
	// check for bit pattern 1001 in bits 4-7
	if(	getField(instruction, 4,7) == 0x09 ) 
	{
		// multiply
		return decodeMultiply(address, instruction);
	}
	else
	{
		// data processing
		return decodeDataProcessing(address, instruction);
	}

	return "ERROR DISASSEMBLING IN decodeMultiplyOrDataProcessing()";
}

CString CArmDisassembler::decodeSingleDataSwapOrDataProcessing(uint32 address, uint32 instruction)
{
	// check for bit pattern 0000 1001 in bits 4-11
	if( getField(instruction, 4,11) == 0x09)
	{
		// single data swap
		return decodeSingleDataSwap(address, instruction);
	}
	else
	{
		// data processing
		return decodeDataProcessing(address, instruction);
	}

	return "ERROR DISASSEMBLING IN decodeSingleDataSwapOrDataProcessing()";
}

CString CArmDisassembler::decodeDataProcessing(uint32 address, uint32 instruction)
{
	CString output;
	
	// table of opcode names
	CString opcodeNames[16] =
	{
		"and", "eor", "sub", "rsb",
		"add", "adc", "sbc", "rsc",
		"tst", "teq", "cmp", "cmn",
		"orr", "mov", "bic", "mvn"
	};

	bool useRd[16] =
	{
		TRUE, TRUE, TRUE, TRUE,
		TRUE, TRUE, TRUE, TRUE,
		FALSE, FALSE, FALSE, FALSE,
		TRUE, TRUE, TRUE, TRUE
	};

	bool useRn[16] =
	{
		TRUE, TRUE, TRUE, TRUE,
		TRUE, TRUE, TRUE, TRUE,
		TRUE, TRUE, TRUE, TRUE,
		TRUE, FALSE, TRUE, FALSE
	};

	// deduce opcode for specific data processing instruction
	// based on bits 21-24
	uint8 opcode = (uint8)( getField(instruction, 21,24) );
	output = opcodeNames[ opcode ];

	// add condition code information
	output += decodeConditionCode(instruction);

	// table of which flags don't need explicit S adding
	// so that S is not added for tst, teq, cmp and cmn
	bool opcodeSetsFlagsExplicitly[16] =
	{
		TRUE, TRUE, TRUE, TRUE,
		TRUE, TRUE, TRUE, TRUE,
		FALSE, FALSE, FALSE, FALSE,
		TRUE, TRUE, TRUE, TRUE
	};

	// test S flag (set PSR flags), based on bit 20 and whether this is implicit in the opcode
	if( getBit(instruction, 20) && opcodeSetsFlagsExplicitly[opcode] )
	{
		output += "S";
	}

	// append spaces as necessary, so mnemonic with suffixes is 8 characters long
	while( output.GetLength() < 8 )
	{
			output += " ";
	}
	
	CString registerNumber;

	if(useRd[getField(instruction, 21,24)] )
	{
		// add destination register
		CString registerNumber;
		registerNumber.Format("%d", (instruction >> 12) & 0x0F );
		output += "r" + registerNumber;
	}

	// the 'mov' and 'mvn' instructions dosn't use the first operand so don't output it
	if( useRn[getField(instruction, 21,24)] )
	{

		// add comma if needed
		if( useRd[getField(instruction, 21, 24)] )
			output += ",";

		// add first operand register
		registerNumber.Format("%d", getField(instruction, 16, 19) );
		output += "r" + registerNumber;
	}

	// add comma
	output += ",";

	// decide type of second operand from bit 25
	if( getBit(instruction, 25) )
	{
		// 2nd operand is immediate
		
		// get immediate value from bits 0-7
		uint32 immediate = getField(instruction, 0,7);

		// get rotate (right) applied to immediate from bits 8-11
		uint32 rotate = getField(instruction, 8,11);
		rotate *= 2;

		// rotate immediate right
		uint32 actualImmediate = (immediate >> rotate) | (immediate << (32 - rotate));

		// add immediate value
		CString immediateValue;
		immediateValue.Format("0x%x", actualImmediate);
		output += "#" + immediateValue;
	}
	else
	{
		// 2nd operand is register

		// add second operand register (to which the shift is applied) from bits 0-3
		registerNumber.Format("%d", instruction & 0x0F );
		output += "r" + registerNumber;		

		// if there's any shift at all
		if( getField(instruction, 4, 11) )
		{
			output += ",";

			// table of shift mnemonics
			CString shiftMnemonic[4] =
			{
				"lsl", "lsr", "asr", "ror"
			};

			// ??? cope with RRX

			// ??? cope with LSR #0 encoding being LSR #32

			// add shift mnemonic from bits 5-6
			output += shiftMnemonic[ getField(instruction, 5,6) ] + " ";

			// decide type of shift amount based on bit 4
			if( getBit(instruction, 4) )
			{
				// shifted by amount in register
			
				// add register to shift by from bits 8-11
				registerNumber.Format("%d", getField(instruction, 8,11) );
				output += "r" + registerNumber;
			}
			else
			{
				// shifted by amount in immediate

				// add immediate value to shift by from bits 7-11
				output += "#";
				CString immediateValue;
				immediateValue.Format("%d", getField(instruction, 7,11) );
				output += immediateValue;
			}
		} // end of if any shift
	}

	return output;
}

CString CArmDisassembler::decodeSingleDTImmOffsetPostIndex(uint32 address, uint32 instruction)
{
	
	CString output;

	// decode whether load or store from bit 20
	if( getBit(instruction, 20) )
	{
		// load from memory
		output = "ldr";
	}
	else
	{
		// store to memory
		output = "str";
	}

	// add condition code information
	output += decodeConditionCode(instruction);

	// decode whether we're dealing in byte quantities or not from bit 22
	if( getBit(instruction, 22) )
	{
		// byte quantity
		output += "b";
	}

	// check for trans bit
	if( getBit(instruction, 21) )
        output += "t";

	// append spaces as necessary, so mnemonic with suffixes is 8 characters long
	while( output.GetLength() < 8 )
	{
			output += " ";
	}

	CString registerNumber;
	
	// add source/destination register from bits 12-15
	registerNumber.Format("%d", getField(instruction, 12,15) );
	output += "r" + registerNumber;

	output += ",[";

	// add base register from bits 16-19
	registerNumber.Format("%d", getField(instruction, 16,19) );
	output += "r" + registerNumber;

	output += "],#";

	// add up/down sign from bit 23
	if( !getBit(instruction, 23) )
	{
		// if down then add subtract symbol
		output += "-";
	}

	// add unsigned 12 bit immediate offset from bits 0-11
	CString immediateOffset;
	immediateOffset.Format("%d", getField(instruction, 0,11) );
	output += immediateOffset;

	return output;
}

CString CArmDisassembler::decodeSingleDTImmOffsetPreIndex(uint32 address, uint32 instruction)
{
	CString output;

	// decode whether load or store from bit 20
	if( getBit(instruction, 20) )
	{
		// load from memory
		output = "ldr";
	}
	else
	{
		// store to memory
		output = "str";
	}

	// add condition code information
	output += decodeConditionCode(instruction);

	// decode whether we're dealing in byte quantities or not from bit 22
	if( getBit(instruction, 22) )
	{
		// byte quantity
		output += "b";
	}

	// append spaces as necessary, so mnemonic with suffixes is 8 characters long
	while( output.GetLength() < 8 )
	{
			output += " ";
	}

	CString registerNumber;
	
	// add source/destination register from bits 12-15
	registerNumber.Format("%d", getField(instruction, 12,15) );
	output += "r" + registerNumber;

	output += ",[";

	// add base register from bits 16-19
	registerNumber.Format("%d", getField(instruction, 16,19) );
	output += "r" + registerNumber;

	output += ",#";

	// add up/down sign from bit 23
	if( !getBit(instruction, 23) )
	{
		// if down then add subtract symbol
		output += "-";
	}

	// add unsigned 12 bit immediate offset from bits 0-11
	CString immediateOffset;
	immediateOffset.Format("%d", getField(instruction, 0,11) );
	output += immediateOffset;

	output += "]";

	// decode write-back from bit 21
	if( getBit(instruction, 21) )
	{
		// write-back present
		output += "!";
	}

	return output;
}

CString CArmDisassembler::decodeSingleDTRegOffsetPostIndex(uint32 address, uint32 instruction)
{
	CString output;

	// decode whether load or store from bit 20
	if( getBit(instruction, 20) )
	{
		// load from memory
		output = "ldr";
	}
	else
	{
		// store to memory
		output = "str";
	}

	// add condition code information
	output += decodeConditionCode(instruction);

	// decode whether we're dealing in byte quantities or not from bit 22
	if( getBit(instruction, 22) )
	{
		// byte quantity
		output += "b";
	}

	// check for trans bit
	if( getBit(instruction, 21) )
        output += "t";

	// append spaces as necessary, so mnemonic with suffixes is 8 characters long
	while( output.GetLength() < 8 )
	{
			output += " ";
	}

	CString registerNumber;
	
	// add source/destination register from bits 12-15
	registerNumber.Format("%d", getField(instruction, 12,15) );
	output += "r" + registerNumber;

	output += ",[";

	// add base register from bits 16-19
	registerNumber.Format("%d", getField(instruction, 16,19) );
	output += "r" + registerNumber;

	output += "],";

	// ??? it would seem perverse to have an up/down bit apply to a signed register
	// however this needs to be confirmed.

	// add offset register from bits 0-3
	registerNumber.Format("%d", getField(instruction, 0,3) );
	output += "r" + registerNumber;

	uint32 imm = getField(instruction, 7,11);

	// table of shift mnemonics
	CString shiftMnemonic[4] =
	{
		"lsl", "lsr", "asr", "ror"
	};

	CString shiftType = shiftMnemonic[ getField(instruction, 5,6) ];

	// if not LSL #0
	if( !(shiftType == "lsl" && imm ==0) )
	{

		output += ",";

		// add shift mnemonic from bits 5-6
		output += shiftType + " ";

		// note, single data stores can only be shifted by an immediate amount

		// add immediate value to shift by from bits 7-11
		output += "#";
		CString immediateValue;
		immediateValue.Format("%d", imm );
		output += immediateValue;

	}
	return output;
}

CString CArmDisassembler::decodeSingleDTRegOffsetPreIndex(uint32 address, uint32 instruction)
{
	CString output;

	// decode whether load or store from bit 20
	if( getBit(instruction, 20) )
	{
		// load from memory
		output = "ldr";
	}
	else
	{
		// store to memory
		output = "str";
	}

	// add condition code information
	output += decodeConditionCode(instruction);

	// decode whether we're dealing in byte quantities or not from bit 22
	if( getBit(instruction, 22) )
	{
		// byte quantity
		output += "b";
	}
	
	// append spaces as necessary, so mnemonic with suffixes is 8 characters long
	while( output.GetLength() < 8 )
	{
			output += " ";
	}

	CString registerNumber;
	
	// add source/destination register from bits 12-15
	registerNumber.Format("%d", getField(instruction, 12,15) );
	output += "r" + registerNumber;

	output += ",[";

	// add base register from bits 16-19
	registerNumber.Format("%d", getField(instruction, 16,19) );
	output += "r" + registerNumber;

	output += ",";

	// ??? it would seem perverse to have an up/down bit apply to a signed register
	// however this needs to be confirmed.

	// add offset register from bits 0-3
	registerNumber.Format("%d", getField(instruction, 0,3) );
	output += "r" + registerNumber;

	uint32 imm = getField(instruction, 7,11);

	// table of shift mnemonics
	CString shiftMnemonic[4] =
	{
		"lsl", "lsr", "asr", "ror"
	};

	CString shiftType = shiftMnemonic[ getField(instruction, 5,6) ];

	// if not LSL #0
	if( !(shiftType == "lsl" && imm == 0) )
	{

		output += ",";

		// add shift mnemonic from bits 5-6
		output +=  shiftType + " ";

		// note, single data stores can only be shifted by an immediate amount

		// add immediate value to shift by from bits 7-11
		output += "#";
		CString immediateValue;
		immediateValue.Format("%d", imm );
		output += immediateValue;

	}

	output += "]";

	// check for write-back from bit 21
	if( getBit(instruction, 21) )
	{
		// write-back
		output += "!";
	}

	return output;
}

CString CArmDisassembler::decodeBlockDTPostIndex(uint32 address, uint32 instruction)
{
	
	CString output;

	// decode whether to load from or store to memory from bit 20
	if( getBit(instruction, 20) )
	{
		// load from memory
		output = "ldm";
	}
	else
	{
		// store to memory
		output = "stm";
	}

	// add condition code information
	output += decodeConditionCode(instruction);

	// decode whether to increment or decrement base from bit 23
	if( getBit(instruction, 23) )
	{
		// increment
		output += "i";
	}
	else
	{
		// decrement
		output += "d";
	}

	// we already decoded from the initial lookup that it's post indexed
	output += "a";

	// append spaces as necessary, so mnemonic with suffixes is 8 characters long
	while( output.GetLength() < 8 )
	{
		output += " ";
	}

	CString registerNumber;

	// decode base register from bits 16-19
	registerNumber.Format("%d", getField(instruction, 16,19) );
	output += "r" + registerNumber;

	// decode write-back from bit 21
	if( getBit(instruction, 21) )
	{
		output += "!";
	}

	output += ",{";

	// decode register list from bits 0-15
	output += decodeRegisterList(instruction);

	// decode whether to load PSR and force user mode or not from bit 22
	if( getBit(instruction, 22) )
	{
		output += "^";
	}

	return output;
}

CString CArmDisassembler::decodeBlockDTPreIndex(uint32 address, uint32 instruction)
{
	CString output;

	// decode whether to load from or store to memory from bit 20
	if( getBit(instruction, 20) )
	{
		// load from memory
		output = "ldm";
	}
	else
	{
		// store to memory
		output = "stm";
	}

	// add condition code information
	output += decodeConditionCode(instruction);

	// decode whether to increment or decrement base from bit 23
	if( getBit(instruction, 23) )
	{
		// increment
		output += "i";
	}
	else
	{
		// decrement
		output += "d";
	}

	// we already decoded from the initial lookup that it's pre indexed
	output += "b";

	// append spaces as necessary, so mnemonic with suffixes is 8 characters long
	while( output.GetLength() < 8 )
	{
		output += " ";
	}

	CString registerNumber;

	// decode base register from bits 16-19
	registerNumber.Format("%d", getField(instruction, 16,19) );
	output += "r" + registerNumber;

	// decode write-back from bit 21
	if( getBit(instruction, 21) )
	{
		output += "!";
	}

	output += ",{";

	// decode register list from bits 0-15
	output += decodeRegisterList(instruction);

	// decode whether to load PSR and force user mode or not from bit 22
	if( getBit(instruction, 22) )
	{
		output += "^";
	}

	return output;
}

CString CArmDisassembler::decodeRegisterList(uint32 instruction)
{
	CString registerList;
	uint8 run = 0;
	uint8 previousRegisterSelected = 99;	// or any invalid register number
	CString registerNumber;

	BOOL commaNeeded = FALSE;	// comman not needed

	// for each of the 16 registers, r0 to r15
	for(uint8 regNumber=0; regNumber<16; regNumber++)
	{

		if(run)
		{
			if(getBit(instruction, regNumber) )
			{
				// bit set
				if( regNumber<15 && getBit(instruction, regNumber+1) )
				{
					// if we're second element in the run add - 
					if(run == 1)
						registerList +="-";
					run++;
				}
				else
				{
					// if run was only 1 reg then need to add a comma
					if(run == 1)
						registerList += ",";
					
					// next bit not set or dealing with r15
					// we're at the end of a list
					registerNumber.Format("%d",regNumber);
					registerList += "r" + registerNumber;

					commaNeeded = TRUE;
					run = 0;
				}
			}
			else
			{
				// last reg so end list
				registerNumber.Format("%d",regNumber);
				registerList += "r" + registerNumber;
				run = 0;
				commaNeeded =TRUE;
			}
		}
		else
		{
			// not in run
			if( getBit(instruction, regNumber) )
			{
				if( regNumber<15 && getBit(instruction, regNumber+1) )
				{
					// if next bit set, then we're first of run
					
					// place comma if needed
					if(commaNeeded)
						registerList += ",";
					
					registerNumber.Format("%d",regNumber);
					registerList += "r" + registerNumber;

					run++;
				}
				else
				{
					if(commaNeeded)
						registerList +=",";
					
					// next bit not set or we're looking at r15 (end list)
					registerNumber.Format("%d",regNumber);
					registerList += "r" + registerNumber;
					
					run = 0;
					commaNeeded = TRUE;
				}
			}
		}
	}

	registerList += "}";

	return registerList;
}

// ??? absolute address generated is wrong
CString CArmDisassembler::decodeBranch(uint32 address, uint32 instruction)
{
	CString output;

	output = "b";

	output += decodeConditionCode(instruction);

	// append spaces as necessary, so mnemonic with suffixes is 8 characters long
	while( output.GetLength() < 8 )
	{
		output += " ";
	}

	// get offset from bits 0-23
	// note how shifted left 2 as instructions are aligned to 32 bit word boundaries.
	uint32 offset = getField(instruction, 0,23);
	offset <<= 2;

	// add offset to instruction's address and add 8 for the pipelining effect
	// i.e. PC is 8 bytes ahead of the currently executing instruction
	uint32 branchAddress = address + offset + 8;
	// limit size to 26 bit address
	branchAddress &= (1<<26)-1;
	
	// output branch address in hex
	CString branchAddressHex;
	branchAddressHex.Format("0x%x", branchAddress);
	output += branchAddressHex;

	return output;
}

CString CArmDisassembler::decodeBranchWithLink(uint32 address, uint32 instruction)
{
	CString output;

	output = "bl";

	output += decodeConditionCode(instruction);

	// append spaces as necessary, so mnemonic with suffixes is 8 characters long
	while( output.GetLength() < 8 )
	{
		output += " ";
	}

	// get offset from bits 0-23
	// note how shifted left 2 as instructions are aligned to 32 bit word boundaries.
	uint32 offset = getField(instruction, 0,23);
	offset <<= 2;

	// add offset to instruction's address and add 8 for the pipelining effect
	// i.e. PC is 8 bytes ahead of the currently executing instruction
	uint32 branchAddress = address + offset + 8;
	// limit size to 26 bit address
	branchAddress &= (1<<26)-1;
	
	// output branch address in hex
	CString branchAddressHex;
	branchAddressHex.Format("0x%x", branchAddress);
	output += branchAddressHex;

	return output;
}

CString CArmDisassembler::decodeCoProDTPreIndex(uint32 address, uint32 instruction)
{
	return "CO PRO DATA TRANSFER PRE INDEX";
}

CString CArmDisassembler::decodeCoProDTPostIndex(uint32 address, uint32 instruction)
{
	return "CO PRO DATA TRANSFER POST INDEX";
}

CString CArmDisassembler::decodeCoProRegTransferOrDataOperation(uint32 address, uint32 instruction)
{
	return "CO PRO REG TRANSFER OR DATA OP";
}

CString CArmDisassembler::decodeSoftwareInterrupt(uint32 address, uint32 instruction)
{
	CString output = "swi" + decodeConditionCode(instruction);
	
	// append spaces as necessary, so mnemonic with suffixes is 8 characters long
	while( output.GetLength() < 8 )
	{
			output += " ";
	}

	CString swiNumber;
	swiNumber.Format("0x%x", getField(instruction, 0,23) );
	output += swiNumber;

	return output;
}

CString CArmDisassembler::decodeMultiply(uint32 address, uint32 instruction)
{
	CString output;
	bool accumulate = FALSE;

	// decide whether additional accumulate function from bit 21
	if( getBit(instruction, 21) )
	{
		// multiply and accumulate
		output = "mla";
		accumulate = TRUE;		
	}
	else
	{
		// standard multiply
		output = "mul";
	}

	output += decodeConditionCode(instruction);

	// test S flag (set PSR flags), based on bit 20
	if( getBit(instruction, 20) )
	{
		output += "S";
	}

	// append spaces as necessary, so mnemonic with suffixes is 8 characters long
	while( output.GetLength() < 8 )
	{
			output += " ";
	}

	CString registerNumber;

	// add destination register from bits 16-19
	registerNumber.Format("%d", getField(instruction, 16,19) );
	output += "r" + registerNumber;

	output += ",";

	// add 1st operand register from bits 0-3
	registerNumber.Format("%d", getField(instruction, 0,3) );
	output += "r" + registerNumber;

	output += ",";

	// add 2nd operand register from bits 8-11
	registerNumber.Format("%d", getField(instruction, 8,11) );
	output += "r" + registerNumber;

	if(accumulate)
	{
		output += ",";

		// add 3rd operand register from bits 12-15
		registerNumber.Format("%d", getField(instruction, 12,15) );
		output += "r" + registerNumber;
	}

	return output;
}

CString CArmDisassembler::decodeSingleDataSwap(uint32 address, uint32 instruction)
{
	CString output;

	// word or byte quantity, byte if bit 22 set
	if(getBit(instruction,22))
	{
		output = "swpb";
	}
	else
	{
		output = "swp";
	}

	output += decodeConditionCode(instruction);

	while( output.GetLength() < 8 )
	{
		output += " ";
	}

	CString registerNumber;

	// add destination register from bits 12-15
	registerNumber.Format("%d", getField(instruction, 12,15) );
	output += "r" + registerNumber;

	output += ",";

	// add rm from bits 0-3
	registerNumber.Format("%d", getField(instruction, 0, 3) );
	output += "r" + registerNumber;

	output += ",[";

	// add rn from bits 16-19
	registerNumber.Format("%d", getField(instruction, 16, 19) );
	output += "r" + registerNumber;

	output += "]";

	return output;
}

CString CArmDisassembler::decodeConditionCode(uint32 instruction)
{
	// table of condition code meanings, note that as convention dictates, AL is blank
	CString conditionCodes[16] =
	{
		"eq", "ne", "cs", "cc",
		"mi", "pl", "vs", "vc",
		"hi", "ls", "ge", "lt",
		"gt", "le", "", "nv"
	};
	
	return conditionCodes[ getField(instruction, 28,31) ];
}






